#
# Reference :
# Noble Mushtak and Daniel Lemire, Fast Number Parsing Without Fallback (to appear)
#

all_tqs = []

# Generates all possible values of T[q]
# Appendix B of  Number parsing at a gigabyte per second.
# Software: Practice and Experience 2021;51(8):1700–1727.
for q in range(-342, -27):
    power5 = 5 ** -q
    z = 0
    while (1 << z) < power5:
        z += 1
    b = 2 * z + 2 * 64
    c = 2 ** b // power5 + 1
    while c >= (1 << 128):
        c //= 2
    all_tqs.append(c)
for q in range(-27, 0):
    power5 = 5 ** -q
    z = 0
    while (1 << z) < power5:
        z += 1
    b = z + 127
    c = 2 ** b // power5 + 1
    all_tqs.append(c)
for q in range(0, 308 + 1):
    power5 = 5 ** q
    while power5 < (1 << 127):
        power5 *= 2
    while power5 >= (1 << 128):
        power5 //= 2
    all_tqs.append(power5)

# Returns the continued fraction of numer/denom as a list [a0; a1, a2, ..., an]
def continued_fraction(numer, denom):
    # (look at page numbers in top-left, not PDF page numbers)
    cf = []
    while denom != 0:
        quot, rem = divmod(numer, denom)
        cf.append(quot)
        numer, denom = denom, rem
    return cf


# Given a continued fraction [a0; a1, a2, ..., an], returns
# all the convergents of that continued fraction
# as pairs of the form (numer, denom), where numer/denom is
# a convergent of the continued fraction in simple form.
def convergents(cf):
    p_n_minus_2 = 0
    q_n_minus_2 = 1
    p_n_minus_1 = 1
    q_n_minus_1 = 0
    convergents = []
    for a_n in cf:
        p_n = a_n * p_n_minus_1 + p_n_minus_2
        q_n = a_n * q_n_minus_1 + q_n_minus_2
        convergents.append((p_n, q_n))
        p_n_minus_2, q_n_minus_2, p_n_minus_1, q_n_minus_1 = (
            p_n_minus_1,
            q_n_minus_1,
            p_n,
            q_n,
        )
    return convergents


# Enumerate through all the convergents of T[q] / 2^137 with denominators < 2^64
found_solution = False
for j, tq in enumerate(all_tqs):
    for _, w in convergents(continued_fraction(tq, 2 ** 137)):
        if w >= 2 ** 64:
            break
            if (tq * w) % 2 ** 137 > 2 ** 137 - 2 ** 64:
                print(f"SOLUTION: q={j-342} T[q]={tq} w={w}")
                found_solution = True
if not found_solution:
    print("No solutions!")



#!/usr/bin/env python3
########################################################################
# Generates a new release.
########################################################################
import sys
import re
import subprocess
import io
import os
import fileinput

if sys.version_info < (3, 0):
    sys.stdout.write("Sorry, requires Python 3.x or better\n")
    sys.exit(1)


def colored(r, g, b, text):
    return f"\033[38;2;{r};{g};{b}m{text} \033[38;2;255;255;255m"


def extractnumbers(s):
    return tuple(map(int, re.findall(r"(\d+)\.(\d+)\.(\d+)", str(s))[0]))


def toversionstring(major, minor, rev):
    return f"{major}.{minor}.{rev}"


print("Calling git rev-parse --abbrev-ref HEAD")
pipe = subprocess.Popen(
    ["git", "rev-parse", "--abbrev-ref", "HEAD"],
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
)
branchresult = pipe.communicate()[0].decode().strip()

if branchresult != "main":
    print(
        colored(
            255,
            0,
            0,
            f"We recommend that you release on main, you are on '{branchresult}'",
        )
    )

ret = subprocess.call(["git", "remote", "update"])

if ret != 0:
    sys.exit(ret)
print("Calling git log HEAD.. --oneline")
pipe = subprocess.Popen(
    ["git", "log", "HEAD..", "--oneline"],
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
)
uptodateresult = pipe.communicate()[0].decode().strip()

if len(uptodateresult) != 0:
    print(uptodateresult)
    sys.exit(-1)

pipe = subprocess.Popen(
    ["git", "rev-parse", "--show-toplevel"],
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
)
maindir = pipe.communicate()[0].decode().strip()
scriptlocation = os.path.dirname(os.path.abspath(__file__))

print(f"repository: {maindir}")

pipe = subprocess.Popen(
    ["git", "describe", "--abbrev=0", "--tags"],
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
)
versionresult = pipe.communicate()[0].decode().strip()

print(f"last version: {versionresult}")
try:
    currentv = extractnumbers(versionresult)
except:
    currentv = [0, 0, 0]
if len(sys.argv) != 2:
    nextv = (currentv[0], currentv[1], currentv[2] + 1)
    print(f"please specify version number, e.g. {toversionstring(*nextv)}")
    sys.exit(-1)
try:
    newversion = extractnumbers(sys.argv[1])
    print(newversion)
except:
    print(f"can't parse version number {sys.argv[1]}")
    sys.exit(-1)

print("checking that new version is valid")
if newversion[0] != currentv[0]:
    assert newversion[0] == currentv[0] + 1
    assert newversion[1] == 0
    assert newversion[2] == 0
elif newversion[1] != currentv[1]:
    assert newversion[1] == currentv[1] + 1
    assert newversion[2] == 0
else:
    assert newversion[2] == currentv[2] + 1

atleastminor = (currentv[0] != newversion[0]) or (currentv[1] != newversion[1])

newmajorversionstring = str(newversion[0])
newminorversionstring = str(newversion[1])
newpatchversionstring = str(newversion[2])
newversionstring = f"{newversion[0]}.{newversion[1]}.{newversion[2]}"
cmakefile = f"{maindir}{os.sep}CMakeLists.txt"

for line in fileinput.input(cmakefile, inplace=1, backup=".bak"):
    line = re.sub(
        r"project\(fast_float VERSION \d+\.\d+\.\d+ LANGUAGES CXX\)",
        f"project(fast_float VERSION {newversionstring} LANGUAGES CXX)",
        line.rstrip(),
    )
    print(line)

print(f"modified {cmakefile}, a backup was made")

versionfilerel = f"{os.sep}include{os.sep}fast_float{os.sep}float_common.h"
versionfile = f"{maindir}{versionfilerel}"

for line in fileinput.input(versionfile, inplace=1, backup=".bak"):
    line = re.sub(
        r"#define FASTFLOAT_VERSION_MAJOR \d+",
        f"#define FASTFLOAT_VERSION_MAJOR {newmajorversionstring}",
        line.rstrip(),
    )
    line = re.sub(
        r"#define FASTFLOAT_VERSION_MINOR \d+",
        f"#define FASTFLOAT_VERSION_MINOR {newminorversionstring}",
        line.rstrip(),
    )
    line = re.sub(
        r"#define FASTFLOAT_VERSION_PATCH \d+",
        f"#define FASTFLOAT_VERSION_PATCH {newpatchversionstring}",
        line.rstrip(),
    )
    print(line)

print(f"{versionfile} modified")

readmefile = f"{maindir}{os.sep}README.md"

for line in fileinput.input(readmefile, inplace=1, backup=".bak"):
    line = re.sub(
        r"https://github.com/fastfloat/fast_float/releases/download/v(\d+\.\d+\.\d+)/fast_float.h",
        f"https://github.com/fastfloat/fast_float/releases/download/v{newversionstring}/fast_float.h",
        line.rstrip(),
    )
    line = re.sub(
        r"GIT_TAG tags/v(\d+\.\d+\.\d+)",
        f"GIT_TAG tags/v{newversionstring}",
        line.rstrip(),
    )
    line = re.sub(
        r"GIT_TAG v(\d+\.\d+\.\d+)\)", f"GIT_TAG v{newversionstring})", line.rstrip()
    )
    print(line)

print(f"modified {readmefile}, a backup was made")

print("running amalgamate.py")
with open(f"{maindir}{os.sep}fast_float.h", "w") as outfile:
    cp = subprocess.run(
        [f"python3", f"{maindir}{os.sep}script{os.sep}amalgamate.py"], stdout=outfile
    )

if cp.returncode != 0:
    print("Failed to run amalgamate")
else:
    print("amalgamate.py ran successfully")
    print(f"You should upload {maindir}{os.sep}fast_float.h")

print("Please run the tests before issuing a release.\n")
print(
    f'to issue release, enter\n git commit -a && git push && git tag -a v{toversionstring(*newversion)} -m "version {toversionstring(*newversion)}" && git push --tags\n'
)



import sys
from math import floor


def log2(x):
    """returns ceil(log2(x)))"""
    y = 0
    while (1 << y) < x:
        y = y + 1
    return y


for q in range(1, 17 + 1):
    d = 5 ** q
    b = 127 + log2(d)
    t = 2 ** b
    c = t // d + 1
    assert c < 2 ** 128
    assert c >= 2 ** 127
    K = 2 ** 127
    if not (c * K * d <= (K + 1) * t):
        print(q)
        top = floor(t / (c * d - t))
        sys.exit(-1)

for q in range(18, 344 + 1):
    d = 5 ** q
    b = 64 + 2 * log2(d)
    t = 2 ** b
    c = t // d + 1
    assert c > 2 ** (64 + log2(d))
    K = 2 ** 64
    if not (c * K * d <= (K + 1) * t):
        print(q)
        top = floor(t / (c * d - t))
        sys.exit(-1)

print("all good")
